# coding=utf-8
import os
import shutil
import zipfile
from copy import deepcopy
from random import randint
from typing import Dict, Any
from pymsword.xml_template import parse, fill, RE_INSTRUCTION, XMLFragment, XMLRunner
import xml.etree.ElementTree as ET

__all__ = [
    "DocxTemplate", "DocxImageInserter"
]

EXPAND_TAGS = {
    #Container tag to table row
    "row": {"{http://schemas.openxmlformats.org/wordprocessingml/2006/main}tr"},
    #Container tag to table cell (not really usable)
    "cell": {"{http://schemas.openxmlformats.org/wordprocessingml/2006/main}tc"},
    #For list item and paragraph
    "p": {"{http://schemas.openxmlformats.org/wordprocessingml/2006/main}p"},
    "image": {"{http://schemas.openxmlformats.org/wordprocessingml/2006/main}t"},
}

def connect_template_ranges(tree):#type: (ET.ElementTree)->None
    """in WOrd XML, the text is split into runs, and each run can have different formatting.
    This interfers with our template engine, so we need to merge all runs that belong to the same template item.

    To do this:
     - take  paragraph tag
     - find all runs that are inside this tag
     - take their unified text
     - find template tags in the unified run text
     - find runs it belongs to them
     - move template code completely to the first run
    """
    #iterate over all w:p tags
    for p in tree.iterfind(".//{http://schemas.openxmlformats.org/wordprocessingml/2006/main}p"):
        #iterate over all text runs of the paragraph
        runs = []
        parts = []
        character_origins = []
        for irun, run in enumerate(p.iterfind("{http://schemas.openxmlformats.org/wordprocessingml/2006/main}r")):
            text = run.find("{http://schemas.openxmlformats.org/wordprocessingml/2006/main}t")
            if text is not None:
                runs.append(run)
                parts.append(text.text)
                #this can be done more effectively using binary search, but whatever.
                character_origins.extend([irun]*len(text.text))
        if len(runs) <= 1:
            #nothing to merge, only one run
            continue
        unified_text = "".join(parts)

        #find all RE_INSTRUCTION regexp matches in the unified text
        needs_glue = False
        for match in RE_INSTRUCTION.finditer(unified_text):
            match_start = match.start()
            match_end = match.end()
            first_origin = character_origins[match_start]
            runs_in_range = character_origins[match_start:match_end]
            unique_runs = sorted(set(runs_in_range))
            if len(unique_runs) != 1:
                #mark origins of the matched text to be the same
                character_origins[match_start:match_end] = [first_origin]*(match_end-match_start)
                needs_glue = True
        if needs_glue:
            run_start = 0
            for run_index, run in enumerate(runs):
                cur_pos = run_start
                while cur_pos < len(character_origins) and character_origins[cur_pos] == run_index:
                    cur_pos += 1
                part = unified_text[run_start:cur_pos]
                if part != parts[run_index]:
                    if not part:
                        #when text run became completely empty - just remove it completely
                        p.remove(run)
                    else:
                        run.find("{http://schemas.openxmlformats.org/wordprocessingml/2006/main}t").text = part
                run_start = cur_pos
class DocxContentTypes(object):
    def __init__(self, data):
        self.content_types = {}
        self.default_content_types = {}
        self._orig_data = data
        self._modified = False
        self._parse(data)
    def _parse(self, data):
        # Parse the XML file using ElementTree
        tree = ET.fromstring(data)
        self._root = tree
        # Now you can work with the XML tree, for example:
        for elem in tree.iter():
            if elem.tag == '{http://schemas.openxmlformats.org/package/2006/content-types}Override':
                ctype = elem.attrib['ContentType']
                partname = elem.attrib['PartName']
                if partname[:1] == "/":
                    partname = partname[1:]
                self.content_types[partname] = ctype
            elif elem.tag == '{http://schemas.openxmlformats.org/package/2006/content-types}Default':
                extension = elem.attrib['Extension'].lower()
                ctype = elem.attrib['ContentType']
                self.default_content_types[extension] = ctype
    def get(self, path):
        """For the given path in the archive, returns its content type"""
        ext = path.rsplit('.', 1)[-1].lower()
        ctype = self.default_content_types.get(ext)
        #try to load override, if ext enough is not enough
        ctype = self.content_types.get(path, ctype)
        return ctype

    def add_extension(self, ext, ctype):
        if ext in self.default_content_types:
            return
        self.default_content_types[ext] = ctype
        ET.SubElement(self._root, "{http://schemas.openxmlformats.org/package/2006/content-types}Default",
                        attrib={"Extension":ext.lower(), "ContentType":ctype})
        self._modified = True

    def to_xml(self):
        if not self._modified: return self._orig_data
        # Generate the XML string from the ElementTree
        return b'<?xml version="1.0" encoding="UTF-8" standalone="yes"?>' + ET.tostring(self._root, encoding="utf-8")


class DocxRels(object):
    def __init__(self, name, data, _isempty=False): #type: (str, str, bool)->None
        self.name = name
        self.root = ET.fromstring(data)
        self._id2rel = {}
        self._target2rel = {}
        self._last_id = 1
        self._modified = False
        self._orig_data = data
        self._isempty = _isempty

        self._parse_ids()

    @staticmethod
    def create_empty(name):
        return DocxRels(name, """<?xml version="1.0" encoding="UTF-8" standalone="yes"?>
<Relationships xmlns="http://schemas.openxmlformats.org/package/2006/relationships"/>""", _isempty=True)


    def _parse_ids(self):
        for child in self.root.iter():
            if child.tag == "{http://schemas.openxmlformats.org/package/2006/relationships}Relationship":
                rid = child.attrib['Id']
                target = child.attrib['Target']
                self._id2rel[rid] = child
                self._target2rel[target] = child

    def new_id(self): #type: ()->str
        """Create new relation ID"""
        i = self._last_id
        while True:
            newid = "rId{}".format(i)
            i += 1
            if newid in self._id2rel: continue
            self._last_id = i
            return newid

    def new_relationship(self, target, type_):
        if target in self._target2rel: raise ValueError("Already have relationship with target {}".format(target))

        rid = self.new_id()
        child = ET.SubElement(self.root, "{http://schemas.openxmlformats.org/package/2006/relationships}Relationship",
                              attrib = {"Id":rid,
                                        "Type":type_,
                                        "Target":target})
        self._modified = True
        return rid

    def tostr(self):
        if not self._modified:
            if self._isempty:
                return ""
            else:
                return self._orig_data
        else:
            #need to generate XML back
            return b'<?xml version="1.0" encoding="UTF-8" standalone="yes"?>' + ET.tostring(self.root, encoding="utf-8")

class DocxRelsCollection(object):
    def __init__(self):
        self.path2rels = {}

    def __iter__(self):
        """Iterate over all rels in the collection"""
        return iter(self.path2rels.values())

    def __len__(self):
        """Returns number of rels in the collection"""
        return len(self.path2rels)

    def load_rels(self, path, data):
        """Load rels file from the path. WHen target is
           word/document.xml, then rels path is
           word/_rels/document.xml.rels"""
        if path in self.path2rels:
            raise ValueError("Already have rels for path {}".format(path))
        self.path2rels[path] = DocxRels(path, data)

    def get_rels(self, target, create_new=True):
        """Returns relationships for the given target, or None if not found"""
        #first construct the rels path
        splitted = target.split("/")
        fname = splitted.pop()
        splitted.append("_rels")
        splitted.append(fname + ".rels")
        relspath = "/".join(splitted)

        rels = self.path2rels.get(relspath)
        if rels is None:
            if create_new:
                #create new rels
                rels = DocxRels.create_empty(relspath)
                self.path2rels[relspath] = rels
        return rels

class DocxTemplate(object):
    def __init__(self, docx_file):
        #Open the docx as a ZIP archive and read content of the
        self.docx = docx_file
        with zipfile.ZipFile(self.docx, 'r') as z:
            self._read_content(z)
            self._read_rels(z)

    def _read_rels(self, z):
        self.rels_catalog = DocxRelsCollection()
        for info in z.infolist():
            ctype = self.content_types.get(info.filename)
            if ctype == 'application/vnd.openxmlformats-package.relationships+xml':
                #read the rels file
                rels_data = z.read(info.filename)
                self.rels_catalog.load_rels(info.filename, rels_data)

    def _read_content(self, z):
        # Open the DOCX file as a ZIP archive
        # Extract the [Content_Types].xml file from the DOCX file
        content_types_xml = z.read('[Content_Types].xml')
        self.content_types = DocxContentTypes(content_types_xml)


    def generate(self, data, output):
        #copy all files from the source archive to the output, except the main document
        docx = DocxDocument(self)
        docx.generate(data, output)

class DocxDocument(object):
    """Instance of the generated document. Stores state that changes during generation"""
    def __init__(self, template): #type: (DocxTemplate)->None
        self.template = template
        self.media = []
        self.temporary_media_dir = None
        self.rels_catalog = deepcopy(template.rels_catalog) #we will modify them, so make a copy
        self._current_item = None
        self._next_image_index = 1
        self.content_types = deepcopy(template.content_types)


    def _parse_template(self, document):
        etree = ET.fromstring(document)
        connect_template_ranges(etree)
        #parse the template from the main part
        return parse(etree, expansion_markers=EXPAND_TAGS)


    def generate(self, data, output):
        ctype_to_action = {
            "application/vnd.openxmlformats-officedocument.wordprocessingml.header+xml": "template",
            "application/vnd.openxmlformats-officedocument.wordprocessingml.footer+xml": "template",
            "application/vnd.openxmlformats-officedocument.wordprocessingml.document.main+xml": "template",
            'application/vnd.openxmlformats-package.relationships+xml': "skip"
        }
        with zipfile.ZipFile(output, 'w') as zout:
            with zipfile.ZipFile(self.template.docx, 'r') as zin:
                for item in zin.infolist():
                    fdata = zin.read(item.filename)
                    #first get defailt ctype for
                    item_ctype = self.content_types.get(item.filename)
                    action = ctype_to_action.get(item_ctype, "copy")
                    if item.filename.lower() == "[content_types].xml":
                        action = "skip" #we will write it later
                    if action == "skip":
                        pass
                    elif action == "template":
                        self._current_item = item.filename
                        zout.writestr(item.filename, self._generate_item(fdata, data))
                        self._current_item = None
                    elif action == "copy":
                        zout.writestr(item.filename, fdata)
                    else:
                        raise RuntimeError("Bad action for item {}: {}".format(item.filename, action))
            self.write_content_types(zout)
            self.write_media(zout)
            self.write_rels(zout)
        self.cleanup()

    def write_content_types(self, z): #type: (zipfile.ZipFile)->None
        #write the content types file to the ZIP archive
        data = self.content_types.to_xml()
        z.writestr('[Content_Types].xml', data)

    def write_rels(self, z): #type: (zipfile.ZipFile)->None
        for rel in self.rels_catalog:
            #write the rels file to the ZIP archive
            data = rel.tostr()
            if not data:
                continue
            z.writestr(rel.name, data)

    def write_media(self, z):#type: (zipfile.ZipFile)->None
        for media_file, media_name in self.media:
            #write the media file to the ZIP archive
            z.write(media_file, media_name)

    def cleanup(self):
        if self.temporary_media_dir is not None:
            shutil.rmtree(self.temporary_media_dir, ignore_errors=True)

    def _generate_item(self, xmldata, data): #type: (str, Dict[str, Any])->bytes
        template = self._parse_template(xmldata)
        filled = fill(template, data, context=self)
        try:
            del filled.attrib["{http://schemas.openxmlformats.org/markup-compatibility/2006}Ignorable"]
        except KeyError:
            pass
        split_runs_by_newlines(filled)
        self._postprocess_tree(filled)
        return b'<?xml version="1.0" encoding="UTF-8" standalone="yes"?>' + ET.tostring(filled)

    def _postprocess_tree(self, tree):
        #self._replace_image_size_formulas(tree) #disabled due to a bug: incorrectly detected page size
        pass

    #disabled
    def _replace_image_size_formulas(self, tree):
        """We search for tags that have attributes cx and cy, whose values start with "=", and evaluate them.
        Evaluating is only possible when inner width and height of the page are known.
        """
        def make_functions(page_inner_width, page_inner_height):
            #creates 2 functions: =fit_width(w,h,scale_up,scale_down) and =fit_height(w,h,scale_up,scale_down)
            #returrning optimal width and height
            def fit_box(w, h, scale_up=False, scale_down=True):
                #"""Returns the optimal size of the box, fitting into the page"""
                if w < page_inner_width and h < page_inner_height:
                    #both smaller, scaling up if needed
                    if scale_up:
                        ratio = min(float(page_inner_width) / w, float(page_inner_height) / h)
                        w = int(w*ratio)
                        h = int(h*ratio)
                elif w > page_inner_width or h > page_inner_height:
                    #at least one is bigger
                    if scale_down:
                        ratio = min(float(page_inner_width) / w, float(page_inner_height) / h)
                        w = int(w*ratio)
                        h = int(h*ratio)
                return w, h

            def fit_width(*args,**kwargs):
                return fit_box(*args,**kwargs)[0]
            def fit_height(*args, **kwargs):
                return fit_box(*args, **kwargs)[1]
            return {"fit_width": fit_width, "fit_height": fit_height}
        #Now we should walk the document tree, keeping track of the page size in every part
        # Default A4-ish content area: ~6.5 x 8.5 inches → 6.5*914400, 8.5*914400 in EMUs
        EMUS_PER_TWIP = 635  # 1 twip = 1/20 point = 1/1440 inch → 914400/1440=635 EMUs per twip

        to_process = [] #list of triples (node, attribute name, formula) to update
        def process_nodes(inner_width_twips, inner_height_twips):
            inner_width_emus = inner_width_twips * EMUS_PER_TWIP
            inner_height_emus = inner_height_twips * EMUS_PER_TWIP
            functions = make_functions(inner_width_emus, inner_height_emus)
            # now updating nodes to process
            for node, attr, formula in to_process:
                # we have a formula, so we need to evaluate it
                float_value = eval(formula, {}, functions)
                value = str(int(float_value))  # convert to emus
                node.attrib[attr] = value
            # clear the to_process list
            del to_process[:]

        def walk_tree(node):
            # Detect section properties → new page size context
            if node.tag == "{http://schemas.openxmlformats.org/wordprocessingml/2006/main}sectPr":
                page_width = page_height = None
                margin_left = margin_right = margin_top = margin_bottom = 0
                for child in node:
                    if child.tag == "{http://schemas.openxmlformats.org/wordprocessingml/2006/main}pgSz":
                        page_width = int(child.attrib.get("{http://schemas.openxmlformats.org/wordprocessingml/2006/main}w", "0"))
                        page_height = int(child.attrib.get("{http://schemas.openxmlformats.org/wordprocessingml/2006/main}h", "0"))
                    if child.tag == "{http://schemas.openxmlformats.org/wordprocessingml/2006/main}pgMar":
                        margin_left = int(child.attrib.get("{http://schemas.openxmlformats.org/wordprocessingml/2006/main}left", "0"))
                        margin_right = int(child.attrib.get("{http://schemas.openxmlformats.org/wordprocessingml/2006/main}right", "0"))
                        margin_top = int(child.attrib.get("{http://schemas.openxmlformats.org/wordprocessingml/2006/main}top", "0"))
                        margin_bottom = int(child.attrib.get("{http://schemas.openxmlformats.org/wordprocessingml/2006/main}bottom", "0"))
                if page_width is not None and page_height is not None:
                    inner_width_twips = page_width - margin_left - margin_right
                    inner_height_twips = page_height - margin_top - margin_bottom
                    process_nodes(inner_width_twips, inner_height_twips)

            elif node.tag == "{http://schemas.openxmlformats.org/drawingml/2006/wordprocessingDrawing}extent" or \
                node.tag == "{http://schemas.openxmlformats.org/drawingml/2006/main}ext":
                #check if we have cx and cy attributes
                for attr in ("cx", "cy"):
                    if attr in node.attrib:
                        value = node.attrib[attr]
                        if value.startswith("="):
                            formula = value[1:] #remove the "="
                            to_process.append((node, attr, formula))
            for child in node:
                walk_tree(child)
        walk_tree(tree)
        if to_process:
            process_nodes(9026, 13958)  # I found these values in a real document, they are close to A4 size

    def get_media_dir(self):
        if self.temporary_media_dir is not None:
            return self.temporary_media_dir
        import tempfile
        #create temporary media directory
        self.temporary_media_dir = tempfile.mkdtemp(prefix="docxmedia-")
        #also create media subdirectory
        return self.temporary_media_dir

    def image(self, image_file, dpi=96, scale_down=True, scale_up=True, size=None, fit_size=None): #type: (str)->XMLFragment
        """Adds image reference to the media list, and returns XML code that uses it
        scale_down: hen size is not specified, allows to scale image down to fit box size
        scale_up: when size is not specified, allows to scale image up to fit box size
        size: if specified, is a tuple (width, height) in pixels, of exact image size. If height is None or -1, it is detected automatically.
        fit_size: if specified, is a tuple (width, height) in pixels, of the size to fit the image into. Hen fitting, scale_up and scale_down are considered.

        """
        if size is not None and fit_size is not None:
            raise ValueError("Cannot specify both size and fit_size parameters")

        from PIL import Image
        #first get rels for the current item
        im = Image.open(image_file)
        px_width, px_height = im.size

        #convert pixels to EMUs
        emu_per_pixel = 914400.0 / dpi
        emu_width = int(px_width * emu_per_pixel)
        emu_height = int(px_height * emu_per_pixel)


        mediadir = self.get_media_dir()
        ext = os.path.splitext(image_file)[1].lower()
        if not ext: raise ValueError("Image name without extension: {}".format(image_file))
        def get_image_temp_path(ext):
            imagename = "genimage{}{}".format(self._next_image_index, ext)
            self._next_image_index += 1
            return imagename, os.path.join(mediadir, imagename)
        need_copy = True
        if ext in (".jpg", ".jpeg"):
            self.content_types.add_extension("jpeg", "image/jpeg")
            ext = "jpeg"
        elif ext == ".png":
            self.content_types.add_extension("png", "image/png")
        else:
            #convert to PNG using PIL
            self.content_types.add_extension("png", "image/png")
            ext = ".png"
            imagename, image_temp_path = get_image_temp_path(".png")
            im.save(image_temp_path)
            need_copy = False #image is already in the right place

        del im

        if need_copy:
            imagename, image_temp_path = get_image_temp_path(ext)
            #copy the image to the media directory with the new name
            shutil.copy(image_file, image_temp_path)

        path_splitted = self._current_item.split("/")
        path_splitted.pop()
        path_splitted.append("media")
        path_splitted.append(imagename)
        full_image_path = "/".join(path_splitted)
        #create a new relationship for the image
        rels = self.rels_catalog.get_rels(self._current_item, create_new=True)
        relid = rels.new_relationship("media/"+imagename,
                                       "http://schemas.openxmlformats.org/officeDocument/2006/relationships/image")
        self.media.append((image_temp_path, full_image_path))
        #return the XML fragment for the image
        pic_id = randint(1000000, 9999999)
        pic_name = "Image {}".format(pic_id)
        #RIght now we can't write them, because it is too hard to know the size of the page to fit the image
        if size is None and fit_size is None:
            #No fitting, use size as is
            formula_width = str(emu_width)
            formula_height = str(emu_height)
        elif size is not None:
            #use the given size
            w = size[0]
            formula_width = str(int(w * emu_per_pixel))
            h = size[1]
            if h is None or h==-1 or h==0:
                h = w * px_height / px_width
            formula_height = str(int(h * emu_per_pixel))
        elif fit_size is not None:
            fit_w, fit_h = fit_size
            #fit the image to the given size
            scale_w = float(fit_w) / px_width
            scale_h = float(fit_h) / px_height
            scale = min(scale_w, scale_h)
            if px_width < fit_w and px_height < fit_h and scale_up:
                #scale up the image
                w = int(px_width * scale)
                h = int(px_height * scale)
            elif (px_width > fit_w or px_height > fit_h) and scale_down:
                #scale down the image
                w = int(px_width * scale)
                h = int(px_height * scale)
            else:
                w, h = px_width, px_height
            formula_width = str(int(w * emu_per_pixel))
            formula_height = str(int(h * emu_per_pixel))
        else:
            raise RuntimeError("Unreachable code reached")

        return XMLFragment.from_text(""""<ns0:drawing>
          <ns2:inline distB="0" distL="0" distR="0" distT="0" ns3:anchorId="1828AD8F" ns3:editId="64F02F6F">
            <ns2:extent cx="{emu_width}" cy="{emu_height}"/>
            <ns2:docPr id="{pic_id}" name="{pic_name}"/>
            <ns2:cNvGraphicFramePr>
              <ns4:graphicFrameLocks noChangeAspect="1"/>
            </ns2:cNvGraphicFramePr>
            <ns4:graphic>
              <ns4:graphicData uri="http://schemas.openxmlformats.org/drawingml/2006/picture">
                <ns5:pic>
                  <ns5:nvPicPr>
                    <ns5:cNvPr id="{pic_id}" name="{pic_name}"/>
                    <ns5:cNvPicPr/>
                  </ns5:nvPicPr>
                  <ns5:blipFill>
                    <ns4:blip ns6:embed="{relid}">
                      <ns4:extLst>
                        <ns4:ext uri="{{28A0092B-C50C-407E-A947-70E740481C1C}}">
                          <ns7:useLocalDpi val="0"/>
                        </ns4:ext>
                      </ns4:extLst>
                    </ns4:blip>
                    <ns4:stretch>
                      <ns4:fillRect/>
                    </ns4:stretch>
                  </ns5:blipFill>
                  <ns5:spPr>
                    <ns4:xfrm>
                      <ns4:off x="0" y="0"/>
                       <ns4:ext cx="{emu_width}" cy="{emu_height}"/>
                    </ns4:xfrm>
                    <ns4:prstGeom prst="rect">
                      <ns4:avLst/>
                    </ns4:prstGeom>
                  </ns5:spPr>
                </ns5:pic>
              </ns4:graphicData>
            </ns4:graphic>
          </ns2:inline>
        </ns0:drawing>""".format(relid=relid, pic_id=pic_id, pic_name=pic_name, emu_width=formula_width, emu_height=formula_height),
            #xmlns:ns0="http://schemas.openxmlformats.org/wordprocessingml/2006/main"
            #xmlns:ns1="http://schemas.microsoft.com/office/word/2010/wordml" xmlns:ns2="http://schemas.openxmlformats.org/drawingml/2006/wordprocessingDrawing" xmlns:ns3="http://schemas.microsoft.com/office/word/2010/wordprocessingDrawing" xmlns:ns4="http://schemas.openxmlformats.org/drawingml/2006/main" xmlns:ns5="http://schemas.openxmlformats.org/drawingml/2006/picture" xmlns:ns6="http://schemas.openxmlformats.org/officeDocument/2006/relationships" xmlns:ns7="http://schemas.microsoft.com/office/drawing/2010/main"
            nsmap={"ns0": "http://schemas.openxmlformats.org/wordprocessingml/2006/main",
                   "ns1": "http://schemas.microsoft.com/office/word/2010/wordml",
                   "ns2": "http://schemas.openxmlformats.org/drawingml/2006/wordprocessingDrawing",
                   "ns3": "http://schemas.microsoft.com/office/word/2010/wordprocessingDrawing",
                   "ns4": "http://schemas.openxmlformats.org/drawingml/2006/main",
                   "ns5": "http://schemas.openxmlformats.org/drawingml/2006/picture",
                   "ns6": "http://schemas.openxmlformats.org/officeDocument/2006/relationships",
                   "ns7": "http://schemas.microsoft.com/office/drawing/2010/main"},
                                     parent_tag = "{http://schemas.openxmlformats.org/wordprocessingml/2006/main}r",)



class DocxImageInserter(XMLRunner):
    def __init__(self, image_path, **kwargs):
        """FOr arguments, see DocxDocument.image() method"""
        self.image_path = image_path
        self.kwargs = kwargs

    def run(self, context):
        assert isinstance(context, DocxDocument)
        return context.image(self.image_path, **self.kwargs)

def split_runs_by_newlines(tree): #type: (ET.Element)->None
    """Find text runs that contain newlines and split them, inserting w:br tags in between"""
    #find all text runs
    for run in tree.iterfind(".//{http://schemas.openxmlformats.org/wordprocessingml/2006/main}r"):
        #get text
        text_tag = run.find("{http://schemas.openxmlformats.org/wordprocessingml/2006/main}t")
        if text_tag is None:
            continue
        text = text_tag.text
        if text is None:
            continue
        parts = text.split("\n")
        if len(parts) > 1:
            text_tag.text = parts[0]
            for part in parts[1:]:
                ET.SubElement(run, "{http://schemas.openxmlformats.org/wordprocessingml/2006/main}br")
                #create text node
                textnode = ET.SubElement(run, "{http://schemas.openxmlformats.org/wordprocessingml/2006/main}t",
                                         attrib={"{http://www.w3.org/XML/1998/namespace}space": "preserve"})
                textnode.text = part




import re
from ast import literal_eval
from copy import deepcopy
import xml.etree.ElementTree as ET
from collections import defaultdict
from itertools import groupby
from typing import Optional, List, Dict, Iterable, Any, Set, Callable, Union


class TemplateElementBase(object):
    def __init__(self, parent): #type: (Optional[TemplateContainerBase])->None
        self.parent = parent
        self.valid = True

    def index(self):
        if self.parent is None: raise ValueError("Root node has no index")
        return self.parent.children.index(self)

    def pretty_print(self, indent=""):
        print(indent, str(self))

    def walk(self): #type: ()->Iterable[TemplateElementBase]
        stack = [self]
        while stack:
            node = stack.pop()
            yield node
            if isinstance(node, TemplateContainerBase):
                stack.extend(reversed(node.children))

    def is_static_data(self):
        """Returns True if this element is static data, i.e. it does not contain any placeholders or groups"""
        return all(isinstance(child, (TemplateText, TemplateNode)) for child in self.walk())

class TemplateContainerBase(TemplateElementBase):
    def __init__(self, parent):
        TemplateElementBase.__init__(self, parent)
        self.children = [] #type: List[TemplateElementBase]
    def pretty_print(self, indent=""):
        print(indent, str(self))
        for child in self.children:
            child.pretty_print(indent + "  ")
    def set_children(self, children):
        self.children = children
        for child in children:
            child.parent = self
    def extend_children(self, children):
        for child in children:
            self.children.append(child)
            child.parent = self

class TemplateNode(TemplateContainerBase):
    def __init__(self, xml_node, parent): #type: (ET.Element, Optional[TemplateContainerBase])->None
        TemplateContainerBase.__init__(self, parent)
        self.xml_node = xml_node
        self.sticky_head = False
        self.sticky_tail = False

    def split_by(self, index):
        """Splits the node into two nodes at the given index.
        The first node contains children[:index], the second node contains children[index+1:]
        (this child at index is excluded)
        Split parts are marked as "sticky" so that they can be merged later.
        """
        if self.parent is None: raise ValueError("Cannot split root node")
        head = TemplateNode(self.xml_node, self.parent)
        head.set_children(self.children[:index])
        head.sticky_tail = True
        tail = TemplateNode(self.xml_node, self.parent)
        tail.set_children(self.children[index+1:])
        tail.sticky_head = True
        central_child = self.children[index]
        #reconnect head, center, tail to the parent instead of self
        self_index = self.index()

        new_siblings = self.parent.children[:self_index]# + self.parent.children[self_index+1:]
        new_siblings.extend([head, central_child, tail])
        new_siblings.extend(self.parent.children[self_index+1:])
        self.parent.children = new_siblings
        central_child.parent = self.parent

        self.valid = False

    def __str__(self):
        s = "Node: "+self.xml_node.tag
        if self.sticky_head:
            s += " (sticky head)"
        if self.sticky_tail:
            s += " (sticky tail)"
        if not self.valid:
            s += " (!invalid!)"
        return s

class TemplateNamedElement(TemplateElementBase):
    def __init__(self, name, parent):
        TemplateElementBase.__init__(self, parent)
        self.name = name

    def __str__(self):
        return "{}({})".format(type(self).__name__, self.name)
class TemplatePlaceholder(TemplateNamedElement):
    pass

class TemplateGroupBase(TemplateNamedElement):
    def __init__(self, name, parent):
        TemplateNamedElement.__init__(self, name, parent)
        self.opposite = None #type: Optional[TemplateGroupBase]
    def __str__(self):
        return "{}({}) matched to:{}".format(type(self).__name__, self.name, self.opposite.name)
    def float_up(self):
        self.parent.split_by(self.index())

class CollectedGroupNode(TemplateContainerBase):
    def __init__(self, name, parent):
        TemplateContainerBase.__init__(self, parent)
        self.name = name
    def __str__(self):
        return "Group({})".format(self.name)

class TemplateGroupOpen(TemplateGroupBase):
    def __init__(self, name, parent, expand_to):
        TemplateGroupBase.__init__(self, name, parent)
        self.expand_to = expand_to

class TemplateGroupClose(TemplateGroupBase):
    pass

class TemplateText(TemplateElementBase):
    def __init__(self, text, parent): #type: (str, Optional[TemplateContainerBase])->None
        TemplateElementBase.__init__(self, parent)
        self.text = text
    def __str__(self):
        return "Text: "+repr(self.text)

class XMLFragment(object):
    def __init__(self, data, parent_tag=None):#type: (List[ET.Element], Optional[str])->None
        """Represents a fragment of XML, which can be put instead of a string into the data
        """
        self.data = data
        self.parent_tag = parent_tag  # type: Optional[str]

    @staticmethod
    def from_text(s, nsmap=None, parent_tag=None):  # type: (str, Optonal[Dict[str, str]], Optional[str]) -> XMLFragment
        """Creates an XMLFragment from a string, with optional namespace map."""
        # Build xmlns attributes from nsmap
        ns_attrs = ""
        if nsmap:
            ns_attrs = " " + " ".join(
                'xmlns:{}="{}"'.format(prefix, uri) for prefix, uri in nsmap.items()
            )
        # Wrap the string with root including namespace declarations
        wrapped = "<root{}>{}</root>".format(ns_attrs, s)
        root = ET.fromstring(wrapped)
        #print wrapped
        frag = XMLFragment([child for child in root if isinstance(child, ET.Element)], parent_tag=parent_tag)
        frag.nsmap = nsmap or {}
        return frag
    def __str__(self):
        return "XMLFragment: [{}]".format("".join(ET.tostring(e) for e in self.data))
    def __len__(self):
        return len(self.data)
    def __getitem__(self, index):
        """Returns the i-th element of the fragment"""
        return self.data[index]
    def __iter__(self):
        """Iterates over the elements of the fragment"""
        return iter(self.data)

class TemplateXMLFragment(TemplateElementBase):
    def __init__(self, data, parent):
        #type: (XMLFragment, Optional[TemplateContainerBase])->None
        TemplateElementBase.__init__(self, parent)
        self.data = data
    def __str__(self):
        return  "TemplateXMLFragment: [{}]".format(self.data)

RE_INSTRUCTION = re.compile(r"\{%(.*?)%}|\{\{(.*?)}}")

class XMLRunner(object):
    def __init__(self, callable_): #type: (Optional[Callable[[Any], Union[str, XMLFragment]]])->None
        assert callable(callable_)
        self._callable = callable_

    def __call__(self, context):
        """Runs the XML runner with the given context.
        Should return an XMLFragment or a string.
        """
        return self.run(context)

    def run(self, context):
        return self._callable(context) #type: XMLFragment

def fill_template(template, data, context=None):#type: (TemplateNode, Dict[str, Any], Any)->TemplateNode
    """Fills the template with the given data.
    Data dict must contain string values for placeholders,
    and list-of-dict values for groups.
    context - argument, passed to runners, if they are present.
    """
    def fill(node, filled_parent, data):#type: (TemplateElementBase, Optional[TemplateContainerBase], Dict[str, Any])->List[TemplateElementBase]
        if isinstance(node, TemplateText):
            return [TemplateText(node.text, filled_parent)]
        elif isinstance(node, TemplateXMLFragment):
            raise ValueError("TemplateXMLFragment can only be present in filled data, template is wrong")
        elif isinstance(node, TemplatePlaceholder):
            value = data.get(node.name, '')
            if isinstance(value, XMLRunner):
                #Value is dynamically generated
                value = value(context)
            if isinstance(value, XMLFragment):
                #if value is XMLFragment, then return it as a TemplateXMLFragment
                return [TemplateXMLFragment(value, filled_parent)]
            else:
                return [TemplateText(str(value), filled_parent)]
        elif isinstance(node, (TemplateGroupOpen, TemplateGroupClose)):
            raise ValueError("Cannot fill uncollected groups!")
        elif isinstance(node, CollectedGroupNode):
            #fill the group by repeating its children and exploding its placeholders
            group_data = data.get(node.name)
            if group_data is None: return []#no data for this group
            if not isinstance(group_data, (list, tuple)): raise ValueError("Data for the group {} should be a list or tuple, not {}".format(repr(node.name), type(group_data)))
            generated = []
            for item_data in group_data:
                updated_data = data.copy()
                updated_data.update(item_data)
                for child in node.children:
                    generated.extend(fill(child, filled_parent, updated_data))
            return generated
        elif isinstance(node, TemplateNode):
            #fill the node by filling its children
            filled = TemplateNode(node.xml_node, filled_parent)
            filled.sticky_head = node.sticky_head
            filled.sticky_tail = node.sticky_tail
            generated = []
            for child in node.children:
                generated.extend(fill(child, filled_parent, data))
            filled.set_children(generated)
            return [filled]
        else:
            raise ValueError("Unsupported node type: {}".format(type(node)))

    if isinstance(template, TemplateNode):
        filled_nodes = fill(template, None, data)
        assert len(filled_nodes) == 1 and isinstance(filled_nodes[0], TemplateNode)
        return filled_nodes[0]
    else:
        assert template.is_static_data()
        return template


def parse_template_xml(root, parent): #type: (ET.Element, Optional[TemplateContainerBase])->TemplateNode
    #Each XML node translates to a TemplateNode
    #Each text piece node translates to 1 or more TemplateText and placeholders

    node = TemplateNode(root, parent)
    #First, parse text of the root
    if root.text:
        node.children.extend(parse_text_template(root.text, node))

    #Then, parse children
    for child in root:
        node.children.append(parse_template_xml(child, node))
        if child.tail:
            node.children.extend(parse_text_template(child.tail, node))
    return node

def parse_text_template(text, parent):#type: (str, TemplateContainerBase)->List[TemplateElementBase]
    #parse the text into TemplateText and placeholders
    #find all instructions and split text by them.
    #Each instruction is either a placeholder or a group
    #Each group has a name and a list of instructions

    #find all instructions
    # text between them is transformed to TemplateText
    # instructions are transformed to TemplatePlaceholder or TemplateGroup
    parsed = []
    last_end = 0
    for match in re.finditer(RE_INSTRUCTION, text):
        # Process the string fragment before the match
        fragment = text[last_end:match.start()]
        if fragment:
            parsed.append(TemplateText(fragment, parent))
        # Process the match
        # it is either {% ... %} or {{ ... }}
        if match.group(1):
            instruction = match.group(1).strip()
            if instruction.startswith("end "):
                parsed.append(TemplateGroupClose(instruction[4:], parent))
            else:
                #It is either {% group %} or {% group expand_marker %}
                #try to parse the marker
                instruction_parts = instruction.rsplit(" ", 1)
                expand_to = None
                if len(instruction_parts)==2:
                    expand_to = instruction_parts[1]
                    instruction = instruction_parts[0].strip()
                parsed.append(TemplateGroupOpen(instruction, parent, expand_to))
        else:
            instruction = match.group(2).strip()
            #if instruction is not starting with '"' or "'", then it is a placeholder
            if not instruction.startswith('"') and not instruction.startswith("'"):
                parsed.append(TemplatePlaceholder(instruction, parent))
            else:
                #it is a text
                #evaluate the string literal
                svalue = literal_eval(instruction)
                parsed.append(TemplateText(svalue, parent))
        last_end = match.end()
    # Process the string fragment after the last match
    fragment = text[last_end:]
    if fragment:
        parsed.append(TemplateText(fragment, parent))
    return parsed

#Now find matching open and close group elements
def find_matching_groups(root): #type: (TemplateNode)->None
    name2group = defaultdict(list) #type: Dict[str, List[TemplateGroupOpen]]
    for child in root.walk():
        if isinstance(child, TemplateGroupOpen):
            name2group[child.name].append(child)
        elif isinstance(child, TemplateGroupClose):
            open_groups = name2group.get(child.name)
            if not open_groups:
                raise ValueError("Group {} is closed but not opened".format(child.name))
            open_group = open_groups.pop()
            if not open_groups:
                del name2group[child.name]
            open_group.opposite = child
            child.opposite = open_group
    if name2group:
        raise ValueError("Some groups are not closed: {}".format(", ".join(name2group.keys())))


def expand_group_markers(template, marker2tags): #type: (TemplateNode, Dict[str, Set[str]])->None
    """
    If open group's expand_to is not None, then
    move group open and close markers up by the XML
    hierarchy, until their parent is in the marker2tags[expand_to]

    Needed because in some cases, we can't put textual markers where we want.

    Example:
        <a> <b> <c> {% group exp_a%} {% end group%}</c> </b> </a>
        marker2tags = {"exp_a": {"b"}}

        then after expansion we get:
        <a> {% group exp_a%}<b> <c> </c> </b>{% end group%} </a>
    """
    expansible_nodes = [child for child in template.walk()
                        if isinstance(child, TemplateGroupOpen) \
                            and child.expand_to is not None and \
                            child.expand_to in marker2tags]
    if not expansible_nodes: return

    def pop_tag_up(node_open, wait_tags, direction): #type: (TemplateElementBase, Set[str], int)->None
        #move the open and close markers up
        while node_open.parent is not None:
            last_pop = node_open.parent.xml_node.tag in wait_tags
            parent_node = node_open.parent
            grandparent_node = parent_node.parent
            if grandparent_node is None:
                raise ValueError("Cannot move node {} up, its parent already at the top level".format(node_open))
            #remove the node from its parent
            parent_node.children.remove(node_open)
            #insert it to the grandparent, right before parent_node
            parent_index = parent_node.index()
            grandparent_node.children.insert(parent_index if direction==-1 else parent_index+1, node_open)
            node_open.parent = grandparent_node
            if last_pop: break

    for node_open in expansible_nodes:
        node_close = node_open.opposite
        assert isinstance(node_close, TemplateGroupClose)
        pop_tag_up(node_open, marker2tags[node_open.expand_to], -1)
        pop_tag_up(node_close, marker2tags[node_open.expand_to], 1)


def parents(node):
    while node.parent is not None:
        yield node.parent
        node = node.parent

def last_common_parents(node1, node2): #type: (TemplateElementBase, TemplateElementBase)->TemplateElementBase
    """Find the last common parent of node1 and node2"""
    parents1 = list(parents(node1))
    parents2 = list(parents(node2))
    common = None
    for p1, p2 in zip(parents1[::-1], parents2[::-1]):
        if p1 is p2:
            common = p1
        else:
            break
    assert common is not None #this should never happen, at least root is common
    return common

def push_group_boundaries_up(template):
    """Sometimes group boundaries are inside different tags.
    FOr example, like this:
    <a><b> xxx {% group %} yyy</b> <c> zzz{% end group %}www </c></a>
    In this case, we need to move the boundaries up, splitting the nodes by the boundary, obtaining the following structure:
    <a><b> xxx</b> {% group %} <b>yyy</b> <c> zzz</c> {% end group %}<c>www </c></a>

    This way, both begin and end of the group are inside the same tag <a/>
    Nodes that were split should be marked as "sticky", so that they can be merged later.
    """

    #Find opening tags of the groups
    groups = [child for child in template.walk()
              if isinstance(child, TemplateGroupOpen)]
    assert all(g.opposite is not None for g in groups)
    for group_open in groups:
        group_close = group_open.opposite
        assert isinstance(group_close, TemplateGroupClose)
        #find the last common parent of the group open and close
        common_parent = last_common_parents(group_open, group_close)
        #float open and close tags up to the common parent
        while group_open.parent is not common_parent:
            group_open.float_up()
        while group_close.parent is not common_parent:
            group_close.float_up()
    assert all(g.parent is g.opposite.parent for g in groups)

def collect_groups(template): #type: (TemplateNode)->None
    """Collects all groups into CollectedGroupNode objects"""
    groups = [child for child in template.walk()
                if isinstance(child, TemplateGroupOpen)]
    # replacing ... [open] ... [close] ... with  ... [collected] ...
    for g in groups:
        parent = g.parent
        if not (g.opposite.parent is parent):
            raise ValueError("Node {} and its opposite {} are not in the same parent: {} and {}".format(g, g.opposite, parent, g.opposite.parent))
        open_index = g.index()
        close_index = g.opposite.index()
        collected_group = CollectedGroupNode(g.name, parent)
        collected_group.set_children(parent.children[open_index+1:close_index])
        #replace all children with
        parent.children[open_index:close_index+1] = [collected_group]
        g.valid = False
        g.opposite.valid = False

def glue_texts(tree):
    if isinstance(tree, TemplateNode):
        new_children = []
        had_glues = False
        for is_text, ielements in groupby(tree.children, lambda x: isinstance(x, TemplateText)):
            elements = list(ielements)
            if is_text:
                if len(elements)==1:
                    new_children.append(elements[0])
                else:
                    new_children.append(TemplateText("".join(e.text for e in elements), tree))
                    had_glues = True
            else:
                new_children.extend(elements)
                for elem in elements:
                    glue_texts(elem)
        if had_glues:
            tree.set_children(new_children)

def glue_sticky_elements(template):
    """TempalteNodes that have matching sticky heads and tails AND refer to the same tag
    can be glued together to a single TemplateNode with joined children
    """
    def walk(root):
        modified = False
        #First glue the deeper elements, then the shallower ones
        for child in root.children:
            if isinstance(child, TemplateNode):
                modified = walk(child) or modified
        #Now glue own children
        if isinstance(root, TemplateNode):
            new_children = []
            had_glues = False
            last_node = None #type: Optional[TemplateNode]
            for child in root.children:
                if not isinstance(child, TemplateNode):
                    last_node = None
                    new_children.append(child)
                else:
                    #maybe this node can be glued with the last node?
                    if last_node and last_node.sticky_tail and child.sticky_head and last_node.xml_node is child.xml_node:
                        #glue them!
                        last_node.extend_children(child.children)
                        last_node.sticky_tail = child.sticky_tail
                        had_glues = True
                    else:
                        new_children.append(child)
                        last_node = child
            if had_glues:
                root.set_children(new_children)
            return modified or had_glues

    while walk(template):
        pass

def generate_xml(filled_template):
    """Generate the XML back from the template.
    Node that element references in the TemplateNodes contain data for the template, not for the filled template."""
    assert isinstance(filled_template, TemplateNode)
    #Create root ElementTree element, copying the parameters from filled_template.tag (which is the ElementTree element of the original template)
    parent_map = {} #type: Dict[ET.Element, ET.Element]

    def walk(template_node, xml_parent): #type: (TemplateElementBase, ET.Element)->None
        if isinstance(template_node, TemplateText):
            #if xml_parent has no children - store data to its text
            if len(xml_parent) == 0:
                xml_parent.text = template_node.text
            else:
                #otherwise, put it to the last child's tail
                xml_parent[-1].tail = template_node.text
        elif isinstance(template_node, TemplateXMLFragment):
            #A list of XML tags, copy them to the xml_parent. Create copies!
            # if parent tag is specified, then go up to that parent,
            # and append the XMLFragment to it
            # then go down and recreate the same structure
            if template_node.data:
                tags_up_to_parent = []
                if template_node.data.parent_tag is not None: #we have to go up
                    while xml_parent.tag != template_node.data.parent_tag:
                        tags_up_to_parent.append((xml_parent.tag, xml_parent.attrib))
                        xml_parent = parent_map[xml_parent]
                        if xml_parent is None:
                            raise ValueError("XML fragment was requested to be put under tag {}, but there are no such parents. Parent hierarchy is:" + "\n".join(
                                "{}".format(tag) for tag, _attrs in tags_up_to_parent[::-1]))
                for xml_element in template_node.data:
                    #create a copy of the XML element, with all its attributes and children
                    xml_parent.append(deepcopy(xml_element))
                #now go down to the parent tag, if needed
                for tag, attrs in tags_up_to_parent[::-1]:
                    xml_parent1 = ET.SubElement(xml_parent, tag, attrs)
                    parent_map[xml_parent1] = xml_parent
                    xml_parent = xml_parent1
            if tags_up_to_parent:
                #parent was changed, so need to return it
                return xml_parent

        elif isinstance(template_node, TemplateNode):
            #create sub-element
            xml_node = ET.SubElement(xml_parent, template_node.xml_node.tag, template_node.xml_node.attrib)
            parent_map[xml_node] = xml_parent
            #fill it with children
            for child in template_node.children:
                new_xml_node = walk(child, xml_node)
                if new_xml_node is not None: #happened when walker inserted a TemplateXMLFragment that required going up to the parent tag
                    xml_node = new_xml_node
        else:
            raise ValueError("Unsupported node type: {}".format(type(template_node)))
    root = ET.Element(filled_template.xml_node.tag,
                      filled_template.xml_node.attrib)
    parent_map[root] = None  # root has no parent
    for child in filled_template.children:
        walk(child, root)
    return root

def parse(template_xml, expansion_markers=None):
    template = parse_template_xml(template_xml, None)
    find_matching_groups(template)
    if expansion_markers is not None:
        expand_group_markers(template, expansion_markers)
    #split the template up by group boundaries
    push_group_boundaries_up(template)
    collect_groups(template)
    return template

def fill(template, data, context=None):
    filled = fill_template(template, data, context=context)
    glue_sticky_elements(filled)
    glue_texts(filled)
    return generate_xml(filled)




import unittest
import xml.etree.ElementTree as ET
from pymsword.xml_template import parse, XMLFragment, fill


class TestXMLTemplate(unittest.TestCase):
    def setUp(self):
        self.sample_xml = """
<root>
<document>
<paragraph>
<text>Some {{placeholder}} text</text>
<text>Some {% group %} {{placeholder2}} {% end group %}</text>
<text>A group that contains tags within: {% group1 %} <tag> {{placeholder3}} </tag> {% end group1 %}</text>
<text>A group that overlaps with the paragraph boundary: {% group2 %} <tag> {{placeholder4}} </tag> </text>
</paragraph>
<paragraph>
<text> continue text {% end group2 %}</text>
</paragraph>
</document>
<a> <b> <c>{% egrp marker_a%} {% end egrp %}</c> </b> </a>
</root>
"""
        self.data = {
            'placeholder': 'placeholder value',
            'group': [
                {'placeholder2': 'placeholder2 value 1'},
                {'placeholder2': XMLFragment.from_text("<aa><bb>xxx</bb></aa>")},
            ],
            'group1': [
                {'placeholder3': 'placeholder3 value 1'},
                {'placeholder3': 'placeholder3 value 2'},
            ],
            'group2': [
                {'placeholder4': 'placeholder4 value 1'},
                {'placeholder4': 'placeholder4 value 2'},
            ],
            'egrp': [{}, {}]
        }

    def test_exact_xml_structure(self):
        root = ET.fromstring(self.sample_xml)
        template = parse(root, expansion_markers={'marker_a': {'b'}})
        filled_tree = fill(template, self.data)

        # Build the expected XML tree
        expected_xml = """
<root>
<document>
<paragraph>
<text>Some placeholder value text</text>
<text>Some  placeholder2 value 1  <aa><bb>xxx</bb></aa> </text>
<text>A group that contains tags within:  <tag> placeholder3 value 1 </tag>  <tag> placeholder3 value 2 </tag> </text>
<text>A group that overlaps with the paragraph boundary:  <tag> placeholder4 value 1 </tag> </text>
</paragraph>
<paragraph>
<text> continue text </text></paragraph><paragraph><text> <tag> placeholder4 value 2 </tag> </text>
</paragraph>
<paragraph>
<text> continue text </text>
</paragraph>
</document>
<a> <b> <c> </c> </b><b> <c> </c> </b> </a>
</root>
"""
        expected_tree = ET.fromstring(expected_xml)

        # Compare trees structurally
        def compare_elements(e1, e2):
            self.assertEqual(e1.tag, e2.tag, f"Tag mismatch: {e1.tag} != {e2.tag}")
            self.assertEqual(
                (e1.text or "").strip(),
                (e2.text or "").strip(),
                f"Text mismatch in <{e1.tag}>: {e1.text!r} != {e2.text!r}"
            )
            self.assertEqual(len(e1), len(e2), f"Child count mismatch in <{e1.tag}>")
            for c1, c2 in zip(e1, e2):
                compare_elements(c1, c2)

        print(ET.tostring(filled_tree, encoding='unicode'))
        compare_elements(filled_tree, expected_tree)

if __name__ == "__main__":
    unittest.main()



# coding=utf-8
import os.path
import sys
from typing import Dict, Any, Callable
from pymsword.xml_template import XMLFragment, XMLRunner
from pymsword.docx_template import DocxTemplate

__all__=["DocxComTemplate"]

class DocxComTemplate(DocxTemplate):
    """Version of the DocxTemplate generator that supports COM-based postprocessing"""
    PLACEHODER_TEMPLATE = "EIWOSCVY{:05d}HFBJSYXO"
    PLACEHOLDER_WORD_SEARCH_PATTERN = "EIWOSCVY[0-9]{5}HFBJSYXO"

    def __init__(self, docx_file):
        super(DocxComTemplate, self).__init__(docx_file)
        self.insert_counter = 0

    def prepare_data(self, data):
        """Data can contain string values, lists and inserters.
        We can't run inserters in the DOCX mode, so instead of running them immediately,
        we replace them with placeholder text,
        store inserters in a array, then load the DOCX in COM mode and run all inserters
        """
        placeholder2inserter = {}
        def new_placeholder():
            self.insert_counter += 1
            return self.PLACEHODER_TEMPLATE.format(self.insert_counter)

        def walk_data(data): #type: (Dict[str, Any])->Dict[str, Any]
            prepared = {}
            for key, value in data.items():
                ukey = str(key)
                if isinstance(value, list):
                    mapped_value = [walk_data(item) for item in value]
                elif isinstance(value, XMLRunner):
                    mapped_value = value
                elif callable(value):
                    #store the inserter
                    placeholder = new_placeholder()
                    placeholder2inserter[placeholder] = value
                    mapped_value = placeholder
                elif isinstance(value, XMLFragment):
                    mapped_value = value
                else:
                    mapped_value = str(value)
                prepared[ukey] = mapped_value
            return prepared
        return walk_data(data), placeholder2inserter

    def generate(self, data:Dict[str, Any], output:str, postprocess=None):
        """Generate the DOCX file from the template and data.
        Additional COM-based post-processing can be done.
        :param data: Data to fill the template. A dictionary with string keys and values can be strings, lists, or inserter functions
        :param output: Output file path for the generated DOCX
        :param postprocess: Optional post-processing function that takes the opened Word document as an argument.
        """
        prepared_data, inserters = self.prepare_data(data)
        DocxTemplate.generate(self, prepared_data, output)
        if inserters or postprocess:
            self._run_inserters(output, inserters, postprocess)

    def _run_inserters(self, docx_file, placeholder2inserter, postprocess):
        """Run inserters in the DOCX file"""
        # Initialize the Word Application
        import win32com.client
        word = win32com.client.Dispatch("Word.Application")
        doc = None

        def find_placeholder_in_range(range_obj):
            placeholders_with_ranges = []
            find = range_obj.Find
            find.ClearFormatting()
            find.Text = self.PLACEHOLDER_WORD_SEARCH_PATTERN
            #accept wildcards
            find.MatchWildcards = True
            while True:
                if not find.Execute():
                    break
                found_range = find.Parent.Duplicate
                matched_text = str(found_range.Text)
                placeholders_with_ranges.append((matched_text, found_range))
            return placeholders_with_ranges

        def find_and_process_markers(range_getter):
            #we need getter because for document it is doc.Range(), for headers and footers it is header.Range
            range_obj = range_getter()
            assert not isinstance(range_obj, str)
            #let's do it in 2 passes.
            # Pass 1
            #First, we find all placeholders and hold Ranges for them in a list
            placeholders_with_ranges = find_placeholder_in_range(range_obj)
            # Pass 2
            for matched_text, found_range in placeholders_with_ranges[::-1]:
                try:
                    inserter = placeholder2inserter.get(matched_text)
                    if inserter is not None:
                        inserter(found_range)
                    else:
                        pass
                except Exception as err:
                    print("Error running inserter:", err, file=sys.stderr)
                self.insert_counter += 1
            #pass 3: remove all placeholders, if they are still there
            range_obj = range_getter()
            placeholders_with_ranges = find_placeholder_in_range(range_obj)
            #remove all found ranges
            for matched_text, found_range in placeholders_with_ranges[::-1]:
                try:
                    found_range.Delete()
                except Exception as err:
                    print("Error replacing placeholder:", err, file=sys.stderr)

        try:
            word.Visible = False
            # Open the DOCX file
            word.DisplayAlerts = 0
            word.ScreenUpdating = False
            doc = word.Documents.Open(os.path.abspath(docx_file)) #need to take absolute path, because Word might have different idea about the current folder
            # Search for the marker string
            # Search for the marker string in headers and footers
            if placeholder2inserter:
                for section in doc.Sections:
                    find_and_process_markers(lambda : section.Range)
                    for header in section.Headers:
                        find_and_process_markers(lambda : header.Range)
                    for footer in section.Footers:
                        find_and_process_markers(lambda : footer.Range)
            if postprocess is not None:
                postprocess(doc)

        finally:
            word.ScreenUpdating = True
            if doc is not None:
                doc.Close(SaveChanges=True)
            # Quit Word
            word.Quit()





"""Various inserters that use COM interface to manipulate documsnts."""
import os
from datetime import datetime
from typing import List, Callable


def table_inserter(data:List[List[str]])->Callable[[object], object]:
    """This function creates an "Inserter" - a function that takes Word.Range object
    And inserts a table into the document at the specified location
    :param data: A list of lists, where each inner list represents a row in the table.
    :return: A function that takes a Word.Range object and inserts a table into the document.
    """
    nrows = len(data)
    ncols = 0 if nrows == 0 else len(data[0])

    def inserter(word_range):
        #THis function accetps a Word.Range object and inserts a table into the document
        if nrows == 0 or ncols == 0:
            return
        table = word_range.Document.Tables.Add(word_range,nrows,ncols,1,2) # wdWord9TableBehavior, wdAutoFitWindow
        for i, row in enumerate(data):
            for j, cell in enumerate(row):
                # Set the text of the cell
                table.Cell(i + 1, j + 1).Range.Text = str(cell)
        return

    return inserter


def _insert_document_inline(word_range, filename:str, set_style=None):
    """Inserts content of the RTF file into specified range of document.
    After operation, range points after the inserted RTF."""
    word_range.Collapse()  # wdCollapseStart
    rngEnd = word_range.Duplicate
    rngEnd.InsertParagraph()
    word_range.InsertFile(os.path.abspath(filename), ConfirmConversions=False)
    if set_style is not None:
        # Set the style of the inserted content
        word_range.Style = word_range.Document.Styles(set_style)
    rngEnd.Characters.Last.Delete()
    rngEnd.Characters.Last.Delete()
    word_range.SetRange(rngEnd.Start, rngEnd.End)

    word_range.Collapse(0)  # wdCollapseEnd

    return word_range

def _insert_picture_inline(insert_range, filename):
    """Inserts picture from file <filename> into specified range of document.
    After operation, range points after the inserted picture.
    Function returns a new Word.InlineShape object."""
    pic = insert_range.InlineShapes.AddPicture (os.path.abspath(filename))
    if not pic:
        raise ValueError("AddPicture returned None, check the file path")
    pic_range = pic.Range
    insert_range.SetRange(pic_range.Start, pic_range.End)
    insert_range.Collapse(0) #wdCollapseEnd
    return pic

def image_inserter(picture_path:str)->Callable[[object], object]:
    """Returns a function that inserts an image into the document at the specified range."""
    def insert(word_range):
        if not os.path.exists(picture_path): raise ValueError("Image file was not found")
        _insert_picture_inline(word_range, picture_path)
    return insert

def document_inserter(rtf_path:str, set_style=None)->Callable[[object], object]:
    """Returns a function that inserts RTF content into the document at the specified range."""
    def insert(word_range):
        if not os.path.exists(rtf_path): raise ValueError("RTF file was not found")
        _insert_document_inline(word_range, rtf_path, set_style = set_style)
    return insert


def anchor_inserter(text:str, anchor:str):
    """Returns a function that inserts a bookmark into the document at the specified range.
    :param text: The text to insert at the bookmark.
    :param anchor: The name of the bookmark to create.
    """
    def insert(word_range):
        word_range.Text = text
        bmk = word_range.Document.Bookmarks.Add(anchor, word_range)
        return bmk
    return insert


def heading_inserter(text:str, level:int=1)->Callable[[object], object]:
    """Returns a function that inserts a heading into the document at the specified range.
    :param text: The text of the heading.
    :param level: The heading level (1 for Heading 1, 2 for Heading 2, etc.).
    """
    def insert(word_range):
        word_range.text = text
        word_range.ParagraphFormat.Style = -2 - level
    return insert


def update_document_toc(document):
    """Post-processing function that updates table of content in the document."""
    document.BuiltinDocumentProperties("Creation Date").Value = datetime.now()
    for toc in document.TablesOfContents:
        toc.Update()
